import sys

if sys.version_info >= (3, 7):  # pragma: no cover

    def is_namespace_module(module):
        return module.__file__ is None


else:  # pragma: no cover

    def is_namespace_module(module):
        return getattr(module, "__file__", None) is None



import sys
import time

import pytest
from django.test import SimpleTestCase, override_settings

from django_linear_migrations.apps import check_max_migration_files


class CheckMaxMigrationFilesTests(SimpleTestCase):
    @pytest.fixture(autouse=True)
    def tmp_path_fixture(self, tmp_path):
        migrations_module_name = "migrations" + str(time.time()).replace(".", "")
        self.migrations_dir = tmp_path / migrations_module_name
        self.migrations_dir.mkdir()
        sys.path.insert(0, str(tmp_path))
        try:
            with override_settings(
                MIGRATION_MODULES={"testapp": migrations_module_name}
            ):
                yield
        finally:
            sys.path.pop(0)

    def test_no_migrations_dir(self):
        self.migrations_dir.rmdir()

        result = check_max_migration_files()

        assert result == []

    def test_empty_migrations_dir(self):
        result = check_max_migration_files()

        assert result == []

    def test_non_package(self):
        self.migrations_dir.rmdir()
        self.migrations_dir.with_suffix(".py").touch()

        result = check_max_migration_files()

        assert result == []

    def test_skipped_unspecified_app(self):
        (self.migrations_dir / "__init__.py").touch()

        result = check_max_migration_files(app_configs=[])

        assert result == []

    def test_dlm_E001(self):
        (self.migrations_dir / "__init__.py").touch()

        result = check_max_migration_files()

        assert len(result) == 1
        assert result[0].id == "dlm.E001"
        assert result[0].msg == "testapp's max_migration.txt does not exist."

    def test_dlm_E002(self):
        (self.migrations_dir / "__init__.py").touch()
        (self.migrations_dir / "max_migration.txt").write_text("line1\nline2\n")

        result = check_max_migration_files()

        assert len(result) == 1
        assert result[0].id == "dlm.E002"
        assert result[0].msg == "testapp's max_migration.txt contains multiple lines."

    def test_dlm_E003(self):
        (self.migrations_dir / "__init__.py").touch()
        (self.migrations_dir / "0001_initial.py").touch()
        (self.migrations_dir / "max_migration.txt").write_text("0001_start\n")

        result = check_max_migration_files()

        assert len(result) == 1
        assert result[0].id == "dlm.E003"
        assert result[0].msg == (
            "testapp's max_migration.txt points to non-existent migration"
            + " '0001_start'."
        )

    def test_dlm_E004(self):
        (self.migrations_dir / "__init__.py").touch()
        (self.migrations_dir / "0001_initial.py").touch()
        (self.migrations_dir / "0002_updates.py").touch()
        (self.migrations_dir / "max_migration.txt").write_text("0001_initial\n")

        result = check_max_migration_files()

        assert len(result) == 1
        assert result[0].id == "dlm.E004"
        assert result[0].msg == (
            "testapp's max_migration.txt contains '0001_initial', but the"
            + " latest migration is '0002_updates'."
        )

    def test_okay(self):
        (self.migrations_dir / "__init__.py").touch()
        (self.migrations_dir / "0001_initial.py").touch()
        (self.migrations_dir / "0002_updates.py").touch()
        (self.migrations_dir / "max_migration.txt").write_text("0002_updates\n")

        result = check_max_migration_files()

        assert result == []



import pkgutil
from importlib import import_module, reload
from pathlib import Path

from django.apps import AppConfig, apps
from django.core.checks import Error, Tags, register
from django.db.migrations.loader import MigrationLoader
from django.utils.functional import cached_property

from django_linear_migrations.compat import is_namespace_module


class DjangoMigrationConflictsAppConfig(AppConfig):
    name = "django_linear_migrations"
    verbose_name = "django-linear-migrations"

    def ready(self):
        register(Tags.models)(check_max_migration_files)


def is_first_party_app_config(app_config):
    # Check if it seems to be installed in a virtualenv
    path = Path(app_config.path)
    return "site-packages" not in path.parts


def first_party_app_configs():
    for app_config in apps.get_app_configs():
        if is_first_party_app_config(app_config):
            yield app_config


class MigrationDetails:
    def __init__(self, app_label, do_reload=False):
        self.app_label = app_label

        # Some logic duplicated from MigrationLoader.load_disk, but avoiding
        # loading all migrations since that's relatively slow.
        self.migrations_module_name, _explicit = MigrationLoader.migrations_module(
            app_label
        )
        try:
            self.migrations_module = import_module(self.migrations_module_name)
        except ModuleNotFoundError:
            # Unmigrated app
            self.migrations_module = None
        else:
            if do_reload:
                reload(self.migrations_module)

    @property
    def has_migrations(self):
        return (
            self.migrations_module is not None
            and not is_namespace_module(self.migrations_module)
            # Django ignores non-package migrations modules
            and hasattr(self.migrations_module, "__path__")
        )

    @cached_property
    def dir(self):
        return Path(self.migrations_module.__file__).parent

    @cached_property
    def names(self):
        return {
            name
            for _, name, is_pkg in pkgutil.iter_modules(self.migrations_module.__path__)
            if not is_pkg and name[0] not in "_~"
        }


dlm_E001_msg = "{app_label}'s max_migration.txt does not exist."
dlm_E001_hint = (
    "If you just installed django-linear-migrations, run 'python manage.py"
    + " makemigrations --initialize-max-migrations'. Otherwise, check how it"
    + " has gone missing."
)

dlm_E002_msg = "{app_label}'s max_migration.txt contains multiple lines."
dlm_E002_hint = (
    "This may be the result of a git merge. Fix the file to contain only the"
    + " name of the latest migration."
)

dlm_E003_msg = (
    "{app_label}'s max_migration.txt points to non-existent migration"
    + " '{max_migration_name}'."
)
dlm_E003_hint = "Edit the max_migration.txt to contain the latest migration's name."

dlm_E004_msg = (
    "{app_label}'s max_migration.txt contains '{max_migration_name}',"
    + " but the latest migration is '{real_max_migration_name}'."
)
dlm_E004_hint = (
    "Edit max_migration.txt to contain '{real_max_migration_name}' or rebase"
    + " '{max_migration_name}' to be the latest migration."
)


def check_max_migration_files(*, app_configs=None, **kwargs):
    errors = []
    for app_config in first_party_app_configs():
        # When only checking certain apps, skip the others
        if app_configs is not None and app_config not in app_configs:
            continue
        app_label = app_config.label
        migration_details = MigrationDetails(app_label)

        if not migration_details.has_migrations:
            continue

        max_migration_txt = migration_details.dir / "max_migration.txt"
        if not max_migration_txt.exists():
            errors.append(
                Error(
                    id="dlm.E001",
                    msg=dlm_E001_msg.format(app_label=app_label),
                    hint=dlm_E001_hint,
                )
            )
            continue

        max_migration_txt_lines = max_migration_txt.read_text().strip().splitlines()
        if len(max_migration_txt_lines) > 1:
            errors.append(
                Error(
                    id="dlm.E002",
                    msg=dlm_E002_msg.format(app_label=app_label),
                    hint=dlm_E002_hint,
                )
            )
            continue

        max_migration_name = max_migration_txt_lines[0]
        if max_migration_name not in migration_details.names:
            errors.append(
                Error(
                    id="dlm.E003",
                    msg=dlm_E003_msg.format(
                        app_label=app_label, max_migration_name=max_migration_name
                    ),
                    hint=dlm_E003_hint,
                )
            )
            continue

        real_max_migration_name = max(migration_details.names)
        if max_migration_name != real_max_migration_name:
            errors.append(
                Error(
                    id="dlm.E004",
                    msg=dlm_E004_msg.format(
                        app_label=app_label,
                        max_migration_name=max_migration_name,
                        real_max_migration_name=real_max_migration_name,
                    ),
                    hint=dlm_E004_hint.format(
                        app_label=app_label,
                        max_migration_name=max_migration_name,
                        real_max_migration_name=real_max_migration_name,
                    ),
                )
            )

    return errors



SECRET_KEY = "NOTASECRET"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        "ATOMIC_REQUESTS": True,
    },
}

TIME_ZONE = "UTC"

INSTALLED_APPS = [
    "tests.testapp",
    "django_linear_migrations",
]



import sys

from django.apps import apps
from django.core.management.commands.makemigrations import Command as BaseCommand

from django_linear_migrations.apps import MigrationDetails, first_party_app_configs


class Command(BaseCommand):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--create-max-migration-files",
            action="store_true",
            help=(
                "Generate max_migration.txt files for first-party apps"
                + " that don't have one, and exit."
            ),
        )

    def handle(self, *app_labels, create_max_migration_files, **options):
        if create_max_migration_files:
            self._create_max_migration_files(set(app_labels))
            return
        super().handle(*app_labels, **options)

    def _create_max_migration_files(self, app_labels):
        # Copied check from base command
        app_labels = set(app_labels)
        has_bad_labels = False
        for app_label in app_labels:
            try:
                apps.get_app_config(app_label)
            except LookupError as err:
                self.stderr.write(str(err))
                has_bad_labels = True
        if has_bad_labels:
            sys.exit(2)

        any_created = False
        for app_config in first_party_app_configs():
            if app_labels and app_config.label not in app_labels:
                continue

            migration_details = MigrationDetails(app_config.label)
            max_migration_txt = migration_details.dir / "max_migration.txt"
            if not max_migration_txt.exists():
                max_migration_name = max(migration_details.names)
                max_migration_txt.write_text(max_migration_name + "\n")
                self.stdout.write(
                    "Created max_migration.txt for {}".format(app_config.label)
                )
                any_created = True

        if not any_created:
            self.stdout.write("No max_migration.txt files need creating.")

    def write_migration_files(self, changes):
        super().write_migration_files(changes)
        if self.dry_run:
            return

        first_party_app_labels = {
            app_config.label for app_config in first_party_app_configs()
        }

        for app_label, app_migrations in changes.items():
            if app_label not in first_party_app_labels:
                continue

            # Reload required as we've generated changes
            migration_details = MigrationDetails(app_label, do_reload=True)
            max_migration_name = app_migrations[-1].name
            max_migration_txt = migration_details.dir / "max_migration.txt"
            max_migration_txt.write_text(max_migration_name + "\n")
